class OptionalFeatures:
    async def check_update(self, update, update_type, update_flavor):
        """
        To be implemented by user, called as validator before an update is handled, if the return value is not True the
        update is not queued

        :param update: The update (a dict) received from the Telegram webhook
        :return: True/Other value
        """
        return True



from . import TelegramEndpoint
from .commons import DisableNotification, ParseMode, ReplyToMessageId, _ChatOrInlineMessage
from .reply_markup import ReplyMarkup
from ..multipart_form_data import MultipartEncoder


class FileEndpoint(TelegramEndpoint, DisableNotification, ReplyMarkup, ReplyToMessageId):
    file_type = None
    attributes = []

    def __init__(self, chat_id, file, is_path=True):
        # note: chat_id is mandatory but is not put as argument in the constructor to allow easier caching

        super().__init__()
        self.args = {}

        if is_path is True:
            self.method = 'POST'
            self.multipart_encoder = MultipartEncoder()
            self.multipart_encoder.add_file(self.file_type, file)
            self.headers = {}
        else:
            self.method = 'GET'
            self.args[self.file_type] = file

    def chat_id(self, chat_id):
        self.args['chat_id'] = chat_id
        return self

    def finalize_data(self):
        # call this method after you're sure that you inserted every file you want to upload
        # useful e.g. for adding the thumbnail (when it'll be implemented) but leaving the chat_id dynamic

        if self.method == 'POST':
            boundary, data = self.multipart_encoder.encode()
            self.headers['Content-Type'] = 'multipart/form-data; boundary=%s' % boundary
            self.data = data

    def serialize(self):
        if 'chat_id' not in self.args:
            raise Exception('FileEndpoint without chat_id')

        if self.method == 'POST':
            return self.method, self.endpoint, self.args, self.data, self.headers
        return self.method, self.endpoint, self.args, None, self.headers


class SerializableToInputMedia:
    serialized_fields = []

    def to_input_media(self):

        serialized = {
            'type': self.file_type,
        }

        # todo
        # upload + attach
        # warning: doesn't work
        if self.method == 'POST':
            raise Exception('Must target updated file')
            pass

        serialized['media'] = self.args[self.file_type]

        attributes = self.attributes + self.serialized_fields
        for attribute in attributes:
            if attribute in self.args:
                serialized[attribute] = self.args[attribute]

        return serialized


class Caption:
    def caption(self, caption):
        self.args['caption'] = caption
        return self


# todo
class Thumb:
    pass


class SendPhoto(FileEndpoint, Caption, ParseMode, SerializableToInputMedia):
    file_type = 'photo'
    serialized_fields = ['caption', 'parse_mode']


class SendAudio(FileEndpoint, Caption, ParseMode, Thumb, SerializableToInputMedia):
    file_type = 'audio'
    attributes = ['duration', 'performer', 'title']
    serialized_fields = ['caption', 'parse_mode', 'thumb']


class SendDocument(FileEndpoint, Caption, ParseMode, Thumb, SerializableToInputMedia):
    file_type = 'document'
    serialized_fields = ['caption', 'parse_mode', 'thumb']


class SendVideo(FileEndpoint, Caption, ParseMode, Thumb, SerializableToInputMedia):
    file_type = 'video'
    attributes = ['duration', 'width', 'height']
    serialized_fields = ['caption', 'parse_mode', 'thumb', 'supports_streaming']

    def supports_streaming(self):
        self.args = {
            **self.args,
            'supports_streaming': True
        }
        return self


class SendAnimation(FileEndpoint, Caption, ParseMode, Thumb, SerializableToInputMedia):
    file_type = 'animation'
    attributes = ['duration', 'width', 'height']
    serialized_fields = ['caption', 'parse_mode', 'thumb']


class SendVoice(FileEndpoint, Caption, ParseMode):
    file_type = 'animation'
    attributes = ['duration']


class SendVideoNote(FileEndpoint, Thumb):
    file_type = 'video_note'
    attributes = ['duration', 'length']


class SendMediaGroup(TelegramEndpoint, DisableNotification, ReplyToMessageId):
    def __init__(self, chat_id, media_array):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'media': [x.to_input_media() for x in media_array]
        }


class GetFile(TelegramEndpoint):
    # https://api.telegram.org/file/bot<token>/<file_path>
    def __init__(self, file_id):
        super().__init__()
        self.args = {
            'file_id': file_id
        }


class EditMessageCaption(TelegramEndpoint, Caption, ParseMode, ReplyMarkup, _ChatOrInlineMessage):
    def __init__(self, chat_or_inline_message_id, message_id):
        super().__init__()
        self.__chat_or_inline_message(chat_or_inline_message_id, message_id)


class EditMessageMedia(TelegramEndpoint, ReplyMarkup, _ChatOrInlineMessage):
    def __init__(self, chat_or_inline_message_id, message_id, media):
        super().__init__()
        self.args = {
            'media': media.to_input_media()
        }
        self.__chat_or_inline_message(chat_or_inline_message_id, message_id)



from abc import ABC, abstractmethod


class Trigger(ABC):
    # can define a function named "callback"
    callback = None

    @abstractmethod
    def match(self, update, update_type, update_flavor):
        """Returns True if update matches, any other value if it doesn't. Async supported."""
        pass

    @abstractmethod
    def handle(self, update, update_type, update_flavor):
        """Returns a tuple (url, endpoint, args, data, headers) or None. Async supported."""
        pass



import mimetypes
import random
import string


class MultipartEncoder:
    def __init__(self):
        self.fields = []
        self.files = []

    def add_file(self, field_name, path, encoding=None):
        with open(path, 'rb') as f:
            content = f.read()

        file_name = path.rsplit('/', 1)
        file_name = file_name[1] if len(file_name) == 2 else file_name[0]

        if encoding is None:
            encoding = mimetypes.guess_type(file_name)[0] or 'application/octet-stream'

        self.files.append((field_name, file_name, encoding, content))

    def add_field(self, name, value):
        self.fields.append((name, value))

    def encode(self):
        boundary = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(16))
        _boundary = b'--' + boundary.encode()

        parts = []

        for name, value in self.fields:
            parts.extend(
                [_boundary,
                 b'Content-Disposition: form-data; name="%s"' % name.encode(),
                 b'',
                 value.encode()
                 ]
            )

        for field_name, file_name, content_type, body in self.files:
            parts.extend(
                [_boundary,
                 b'Content-Disposition: form-data; name="%s"; filename="%s"' %
                 (field_name.encode(), file_name.encode()),
                 b'Content-Type: %s' % content_type.encode(),
                 b'',
                 body
                 ]
            )

        parts.append(_boundary + b'--')
        parts.append(b'')
        return boundary, b'\r\n'.join(parts)



import logging
from asyncio import PriorityQueue, get_event_loop, iscoroutinefunction
from urllib.parse import urlencode
from urllib.request import Request, urlopen


class Bot(object):
    update_types = [
        ('message', 'message'),
        ('edited_message', 'message'),
        ('channel_post', 'message'),
        ('edited_channel_post', 'message'),
        ('inline_query', 'inline_query'),
        ('chosen_inline_result', 'chosen_inline_result'),
        ('callback_query', 'callback_query'),
        ('shipping_query', 'shipping_query'),
        ('pre_checkout_query', 'pre_checkout_query')
    ]
    discover_type_and_flavor = False

    def __init__(self, token, json_lib=None, max_handlers=1, loop=None):
        """
        The Class(TM).

        :param token: The Telegram-given token
        """

        self.token = token

        if json_lib is not None:
            self.json = json_lib
        else:
            import json
            self.json = json

        self.max_handlers = max_handlers

        if loop is None:
            self.loop = get_event_loop()
        else:
            self.loop = loop

        self.logger = logging.getLogger('ubot')
        self.triggers = []
        self.ignore_check_triggers = []
        self.update_queue = PriorityQueue(loop=loop)
        self.base_url = 'https://api.telegram.org:443/bot%s/' % token

    async def api_request(self, method, endpoint, data=None, headers=None):
        """
        Wraps the urllib3 request in a more user friendly way, making it async and premitting the base Telegram API url.

        :param method: The HTTP method name (GET, POST, PUT, DELETE, HEAD)
        :param endpoint: A Telegram API endpoint (e.g. sendMessage), omitting
            the first slash
        :param data: Binary data to send with the request
        :param headers: A dict of headers
        :return: The response from the server, in the form of a dict
        """

        url = '%s%s' % (self.base_url, endpoint)

        if data and 'content-length' not in [x.lower() for x in headers.keys()]:
            headers['Content-Length'] = len(data)

        req = Request(url=url, method=method, data=data, headers=headers)
        return urlopen(req)

    async def set_webhook_url(self, url):
        await self.api_request('GET', 'setWebhook?url=%s' % url)

    async def start(self):  # noqa: C901
        """
        Main loop

        >>> import asyncio
        >>> loop = asyncio.get_event_loop()
        >>> bot = Bot('token')
        >>> loop.run_until_complete(asyncio.wait([
        >>>     loop.create_task(bot.start())
        >>>     # other tasks
        >>> ]))
        >>> loop.run_forever()

        or (if you want to use a custom event loop

        >>> import asyncio
        >>> loop = asyncio.new_event_loop()
        >>> bot = Bot('token', loop=loop)
        >>> loop.run_until_complete(asyncio.wait([
        >>>     loop.create_task(bot.start())
        >>>     # other tasks
        >>> ]))
        >>> loop.run_forever()

        note: single instance, multiple threads unsupported (for now), use an instance of the bot for every thread
        """

        while True:
            _, update = await self.update_queue.get()

            # get type and flavor if needed or simply leave them to None
            if self.discover_type_and_flavor is True:
                for _type, flavor in self.update_types:
                    if _type in update:
                        update_type, update_flavor = _type, flavor
                        break
                else:
                    update_type, update_flavor = None, None
            else:
                update_type, update_flavor = None, None

            # check the update if the function is implemented and skip if it's not passed
            if hasattr(self, 'check_update'):
                check_update = self.check_update
                if iscoroutinefunction(check_update):
                    is_check_passed = await check_update(update, update_type, update_flavor)
                else:
                    is_check_passed = check_update(update, update_type, update_flavor)

                if is_check_passed is True:
                    checked_triggers = self.ignore_check_triggers + self.triggers
                else:
                    checked_triggers = self.triggers
            else:
                checked_triggers = self.triggers

            # loop through the triggers to find one that matches
            curr_handlers = 0

            for trigger in checked_triggers:
                if curr_handlers > self.max_handlers:
                    break

                match = trigger.match

                if iscoroutinefunction(match):
                    is_update_matched = await match(update, update_type, update_flavor)
                else:
                    is_update_matched = match(update, update_type, update_flavor)

                if is_update_matched is not True:
                    continue

                curr_handlers += 1

                # if a matching one is found call the corresponding handler
                handle = trigger.handle

                if iscoroutinefunction(handle):
                    endpoint = await handle(update, update_type, update_flavor)
                else:
                    endpoint = handle(update, update_type, update_flavor)

                if endpoint is None:
                    continue

                method, url, args, data, headers = endpoint

                _url = '%s?%s' % (url, urlencode(args))
                response = await self.api_request(method, _url, data, headers)

                # call the callback if existing
                callback = trigger.callback

                if callback is None:
                    continue

                if iscoroutinefunction(callback):
                    await callback(response)
                else:
                    callback(response)

    def push_update(self, update):
        """
        Pushes an update (already json decoded) into the queue.

        :param update: The update to be pushed in the queue
        """

        self.update_queue.put_nowait((update['update_id'], update))

    def trigger(self, requires_type_and_flavor=False, ignore_check=False):
        """
        Decorates a Trigger, inserting it into the bot check list

        :param requires_type_and_flavor: Forces the bot to discover the type and the flavor of an update and to call
            the trigger methods with those as arguments
        :param ignore_check: ignores a possible "check_update" function
        """

        def decorator(trigger):
            if ignore_check is True:
                self.ignore_check_triggers.append(trigger())
            else:
                self.triggers.append(trigger())

            if requires_type_and_flavor is True:
                self.discover_type_and_flavor = True

            return trigger

        return decorator



import datetime

from . import TelegramEndpoint


class KickChatMember(TelegramEndpoint):
    def __init__(self, chat_id, user_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'user_id': user_id
        }

    def until_date(self, date):
        self.args['until_date'] = date
        return self

    def forever(self):
        self.args['until_date'] = (datetime.datetime.utcnow() + datetime.timedelta(days=369)).timestamp()
        return self


class UnbanChatMember(TelegramEndpoint):
    def __init__(self, chat_id, user_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'user_id': user_id
        }


class RestrictChatMember(TelegramEndpoint):
    def __init__(self, chat_id, user_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'user_id': user_id
        }

    def until_date(self, date):
        self.args['until_date'] = date
        return self

    def can_send_messages(self):
        self.args['can_send_messages'] = True
        return self

    def can_send_media_messages(self):
        self.args['can_send_media_messages'] = True
        return self

    def can_send_other_messages(self):
        self.args['can_send_other_messages'] = True
        return self

    def can_add_web_page_previews(self):
        self.args['can_add_web_page_previews'] = True
        return self


class PromoteChatMember(TelegramEndpoint):
    def __init__(self, chat_id, user_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'user_id': user_id
        }

    def can_change_info(self):
        self.args['can_change_info'] = True
        return self

    def can_post_messages(self):
        self.args['can_post_messages'] = True
        return self

    def can_edit_messages(self):
        self.args['can_edit_messages'] = True
        return self

    def can_delete_messages(self):
        self.args['can_delete_messages'] = True
        return self

    def can_invite_users(self):
        self.args['can_invite_users'] = True
        return self

    def can_restrict_members(self):
        self.args['can_restrict_members'] = True
        return self

    def can_pin_messages(self):
        self.args['can_pin_messages'] = True
        return self

    def can_promote_members(self):
        self.args['can_promote_members'] = True
        return self



from . import TelegramEndpoint
from .commons import DisableNotification, ParseMode, ReplyToMessageId, _ChatOrInlineMessage
from .reply_markup import ReplyMarkup


class SendMessage(TelegramEndpoint, DisableNotification, ParseMode, ReplyMarkup, ReplyToMessageId):
    def __init__(self, chat_id, text):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'text': text
        }

    def disable_web_page_overview(self):
        self.args['disable_web_page_overview'] = True
        return self


class EditMessageText(TelegramEndpoint, ParseMode, ReplyMarkup, _ChatOrInlineMessage):
    def __init__(self, chat_or_inline_message_id, message_id, text):
        super().__init__()
        self.args = {
            'text': text
        }
        self.__chat_or_inline_message(chat_or_inline_message_id, message_id)

    def disable_web_page_overview(self):
        self.args['disable_web_page_overview'] = True
        return self



from . import TelegramEndpoint
from .commons import DisableNotification, ReplyToMessageId
from .reply_markup import ReplyMarkup


class SendVenue(TelegramEndpoint, DisableNotification, ReplyMarkup, ReplyToMessageId):
    def __init__(self, chat_id, latitude, longitude, title, address):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'latitude': latitude,
            'longitude': longitude,
            'title': title,
            'address': address  # noqa: S001
        }

    def foursquare_id(self, foursquare_id):
        self.args['foursquare_id'] = foursquare_id
        return self

    def foursquare_type(self, foursquare_type):
        self.args['foursquare_type'] = foursquare_type
        return self



from setuptools import find_packages, setup

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name='ubot',
    version='3.3.0',
    author='Alessandro Cerruti',
    author_email='thereap3r97@gmail.com',
    description='Basic, easily extendable, Telegram bot class',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/strychnide/ubot',
    license='MIT',
    project_urls={
        'Source': 'https://github.com/strychnide/ubot',
        'Issues': 'https://github.com/strychnide/ubot/issues'
    },
    python_requires='>=3.5',
    install_requires=[
        # 'libmediainfo_cffi'
    ],
    extras_require={
        'dev': [
            'flake8',
            'flake8-sorted-keys',
            'flake8-import-order',
            'flake8-quotes',
            'flake8-bugbear',
            'sphinx',
            'sphinx-autodoc-typehints',
            'sphinx_rtd_theme',
            'coverage',
            'twine'
        ]
    },
    packages=find_packages()
)


from .exceptions import InvalidParseMode


class _ChatOrInlineMessage:
    def __chat_or_inline_message(self, chat_or_inline_message_id, message_id):
        if message_id is not None:
            self.args = {
                **self.args,
                'chat_id': chat_or_inline_message_id,
                'message_id': message_id
            }
        else:
            self.args['inline_message_id'] = chat_or_inline_message_id


class ReplyToMessageId:
    def reply_to_message_id(self, message_id):
        self.args['reply_to_message_id'] = message_id
        return self


class DisableNotification:
    def disable_notification(self):
        self.args['disable_notification'] = True
        return self


class ParseMode:
    def parse_mode(self, _parse_mode):
        if _parse_mode not in ['Markdown', 'HTML']:
            raise InvalidParseMode

        self.args['parse_mode'] = _parse_mode
        return self
